
#define WINDOWS_H

/****************************************************************

    Copyright (c) 1996-1997 Fred Pesch

    This program is free software; you can redistribute it 
    and/or modify it under the terms of the GNU General 
    Public License as published by the Free Software 
    Foundation (version 2 of the License).

    This program is distributed in the hope that it will 
    be useful, but WITHOUT ANY WARRANTY; without even the 
    implied warranty of MERCHANTABILITY or FITNESS FOR A 
    PARTICULAR PURPOSE.  See the GNU General Public License 
    for more details.

    For a copy of the GNU General Public License, write to the 
    Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
    MA 02139, USA.

****************************************************************/

#include <Xm/Xm.h>
#include <Xm/MainW.h>
#include <Xm/RowColumn.h>
#include <Xm/CascadeB.h>
#include <Xm/PushB.h>
#include <Xm/Separator.h>
#include <Xm/Text.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <sys/types.h>
#include <signal.h>
#include <unistd.h>

#include "interface.h"
#include "dirtree.h"
#include "config.h"

struct WinNode {
  Window ID;
  int nb_children;
  struct WinNode **children;
};

typedef struct WinNode TWinNode;

unsigned int GlobalCount;

Atom _FOF_SPECIFIC_PROPERTY;
Atom _EXODE_CLIENT_PROPERTY_TYPE;

int NbButtons=0;
int MaxButtons=0;
Widget *Buttons=NULL;
Window *WindowIds=NULL;
unsigned long *Pids=NULL;

/* tile, stack & cascade parameters */

unsigned int RWin_LeftOffset=80;
unsigned int RWin_TopOffset=50;
unsigned int RWin_RightOffset=80;
unsigned int RWin_BottomOffset=50;
unsigned int RWin_XInterOffset=10;
unsigned int RWin_YInterOffset=40;
unsigned int RWin_CascadeXOffset=30;
unsigned int RWin_CascadeYOffset=40;


void ChangeWindowProperty(void)
{
  char exode_sign[128];
  char client_sign[128];
  char windowid[128];
  char pid[128];
  char dir[256];
  int nbletters;
  int i,j;

  char text[256+4*128+128];
 
  _FOF_SPECIFIC_PROPERTY = XInternAtom(XtDisplay(W_App), "_FOF_SPECIFIC_PROPERTY", False);

  _EXODE_CLIENT_PROPERTY_TYPE = XInternAtom(XtDisplay(W_App), "_EXODE_CLIENT_PROPERTY_TYPE", False);

  strcpy(exode_sign,"EXODE");
  strcpy(client_sign,"FOF");
  
  sprintf(windowid,"0x%x",(unsigned int) XtWindow(W_App));
  sprintf(pid,"%lu",(unsigned long int) getpid());

  if(strcmp(ActualDir->dir.pathname,"")==0)
    strcpy(dir, "/");
  else
  if(GetVarInt("WindowMenuTextMaxLength", &nbletters)!= (int)NULL) {
    if(nbletters>=strlen(ActualDir->dir.pathname))
      strcpy(dir, ActualDir->dir.pathname);
    else {
      for(i=0;i< (nbletters >> 1)-2 ; i++)
	dir[i]=ActualDir->dir.pathname[i];
      dir[i++]='.';
      dir[i++]='.';
      dir[i++]='.';
      dir[nbletters]=(char)0;
      for(j=nbletters-1;j>=i;j--)
	dir[j]=ActualDir->dir.pathname[strlen(ActualDir->dir.pathname)+j-nbletters+1];
    }
  } else {
    strncpy(dir, ActualDir->dir.pathname, 255);
    dir[255]=(char)0;
  }
  
  strcpy(text, exode_sign);
  strcat(text, "#");
  strcat(text, client_sign);
  strcat(text, "#");
  strcat(text, windowid);
  strcat(text, "#");
  strcat(text, pid);
  strcat(text, "#");
  strcat(text, dir);


  XChangeProperty(XtDisplay(W_App), XtWindow(W_App), _FOF_SPECIFIC_PROPERTY, _EXODE_CLIENT_PROPERTY_TYPE, 8, PropModeReplace, text, 256+4*128+128);

}

void ChooseClientCB(Widget w_button, XtPointer app_data, XtPointer widget_data)
{
  int num_client;

  num_client=(int) app_data;

  XMapRaised(XtDisplay(W_App), WindowIds[num_client]);
  XWarpPointer(XtDisplay(W_App), None, WindowIds[num_client], 0,0,0,0,0,0);
}

void GetWinTree(Display *display, TWinNode *root)
{
  int status;
  Window *list_wins;
  unsigned int nb_wins;
  Window wroot,wparent;
  int i;

  status =  XQueryTree(display, root->ID, &wroot, &wparent, &list_wins, &nb_wins);
  if(!status) {
    XFree(list_wins);
    return;
  }

  if(nb_wins==0) {
    XFree(list_wins);
    return;
  }

  root->nb_children=nb_wins;
  root->children=(TWinNode **) XtMalloc(sizeof(TWinNode *)*root->nb_children);

  for(i=0;i<root->nb_children;i++) {
    root->children[i]=(TWinNode *) XtMalloc(sizeof(TWinNode));
    root->children[i]->ID=list_wins[i];
    root->children[i]->nb_children=0;
    root->children[i]->children=NULL;
    GetWinTree(display, root->children[i]);
  }

  XFree(list_wins);
}


void ViewWinTree(TWinNode *root, int decal)
{
  int i;

  for(i=0;i<decal;i++) printf(" ");
  printf("Window %#x : %u children \n", (int) root->ID, root->nb_children);
  for(i=0;i<root->nb_children;i++)
    ViewWinTree(root->children[i], decal+2);
}


void DestroyWinTree(TWinNode *root)
{
  int i;

  if(root->nb_children>0) {
    for(i=0;i<root->nb_children;i++)
      XtFree((char *) root->children[i]);

    XtFree((char *) root->children);
  }
  
  XtFree((char *) root);
}

unsigned int CountWinNodes(TWinNode *root)
{
  int i;
  unsigned int count;

  count=0;

  for(i=0;i<root->nb_children;i++)
    count+=CountWinNodes(root->children[i]);

  count+=root->nb_children;

  return count;
}

void WinTreeList(TWinNode *root, Window *wlist)
{
  int i;

  for(i=0;i<root->nb_children;i++)
    WinTreeList(root->children[i], wlist);

  wlist[GlobalCount++]=root->ID;
}

void ActivateMenuCB(Widget w_button, XtPointer app_data, XtPointer widget_data)
{
  
  TWinNode *root;
  Window *list_wins;
  unsigned int nb_wins;
  Widget w_temp;

  Display *display;
  int i;
  int result;
  Atom ret_atom;
  int ret_format;
  unsigned long ret_nb_elements;
  unsigned long ret_left;
  unsigned char *propval;

  char exode_sign[128];
  char client_sign[128];
  char windowid[128];
  char pid[128];
  char dir[256];

  XmString xmbstring;

  Window wwid;
  unsigned long ppid;

  if(NbButtons!=0) {

    for(i=0;i<NbButtons;i++) {
      XtUnmanageChild(Buttons[i]);
    }

      NbButtons=0;
  }
      

  /* query the current display's window list */
  
  display=XtDisplay(W_App);

  root=(TWinNode *) malloc(sizeof(TWinNode));
  root->ID=DefaultRootWindow(display);
  root->nb_children=0;
  root->children=NULL;

  GetWinTree(display, root);

  /*  ViewWinTree(root, 0);*/

  nb_wins=CountWinNodes(root);

  list_wins=(Window *) XtMalloc(sizeof(Window)*nb_wins);
 
  GlobalCount=0;
  WinTreeList(root, list_wins);

  DestroyWinTree(root);

  /* search the interesting clients */

    for(i=0;i<nb_wins;i++) {
      
      result=XGetWindowProperty(display, list_wins[i], _FOF_SPECIFIC_PROPERTY, 0, 128+32, False, _EXODE_CLIENT_PROPERTY_TYPE, &ret_atom, &ret_format, &ret_nb_elements, &ret_left, &propval);

    /* if the window has the property */

      if(result==Success) { 
	if(ret_nb_elements!=0) {

	  sscanf(propval,"%[^#]#%[^#]#%[^#]#%[^#]#%[^#]",exode_sign, client_sign, windowid, pid,dir);
	  sscanf(windowid, "%x", &wwid);
	  sscanf(pid,"%lu", &ppid);

	  WindowIds=(Window*) XtRealloc((char *) WindowIds, sizeof(Window)*(NbButtons+1));
	  WindowIds[NbButtons]=wwid;
	  Pids=(unsigned long *) XtRealloc((char *) Pids, sizeof(unsigned long )*(NbButtons+1));
	  Pids[NbButtons]=ppid;

	  xmbstring=XmStringCreateSimple(dir);

	  if(NbButtons>=MaxButtons) {
	    Buttons=(Widget *) XtRealloc((char *) Buttons, sizeof(Widget)*(NbButtons+1));
	    MaxButtons=NbButtons+1;

	    w_temp = XtNameToWidget(W_MainWindow, "*WindowsMenu");

	    Buttons[NbButtons]= XtVaCreateManagedWidget("ClientBtn", xmPushButtonWidgetClass, w_temp, XmNlabelString, xmbstring, NULL);

            XtAddCallback(Buttons[NbButtons], XmNactivateCallback, ChooseClientCB, (XtPointer) NbButtons);

	  } else {
	    XtManageChild(Buttons[NbButtons]);
	    XtVaSetValues(Buttons[NbButtons], XmNlabelString, xmbstring, NULL);
      }

	  XmStringFree(xmbstring);

	  NbButtons++;
	}
    
    XFree(propval);
      }
    }

    XtFree((char *)list_wins);
    
}

void TileButtonCB(Widget w_button, XtPointer app_data, XtPointer widget_data)
{

  int i;
  int nb_last_col;
  int nb_windows;
  int max_win;
  int nb_row;
  unsigned int rootwidth;
  unsigned int rootheight;
  unsigned int winwidth;
  unsigned int winheight;
  unsigned int xwin;
  unsigned int ywin;
  Display * display;
  Window root;
  unsigned int b_width;
  unsigned int prof;
  int result;

  display = XtDisplay(W_App);

  nb_windows=NbButtons;
  if(nb_windows==0) return;

  /* calculate the number of rows */

  nb_row= (nb_windows + 1) / 2;

  /* calculate the number of columns on the last row (0 or 1) */

  nb_last_col = nb_windows % 2;

  /* windows sizes */

  result=XGetGeometry(display, DefaultRootWindow(display), &root, &xwin, &ywin, &rootwidth, &rootheight, &b_width, &prof);

  if(result==0) return;

  winwidth=( (rootwidth - RWin_LeftOffset - RWin_RightOffset) - ((nb_row - 1)%2)*RWin_XInterOffset) /2;
  winheight=( (rootheight - RWin_TopOffset - RWin_BottomOffset) - (nb_row - 1)*RWin_YInterOffset) / nb_row;

  /* move the windows */

  if(nb_last_col==1)
    max_win=nb_windows-1;
  else max_win=nb_windows;

  xwin=RWin_LeftOffset;
  ywin=RWin_TopOffset;

  for(i=0;i<max_win;i++) {

    XMoveResizeWindow(display, WindowIds[i], xwin, ywin, winwidth, winheight);
    XRaiseWindow(display, WindowIds[i]);

    xwin+= RWin_XInterOffset + winwidth;
    if(i%2==1) {
      xwin=RWin_LeftOffset;
      ywin+=RWin_YInterOffset + winheight;
    }
  }

  if(nb_last_col==1) {
    XMoveResizeWindow(display, WindowIds[i], xwin, ywin, 2*winwidth+RWin_XInterOffset, winheight);
    XRaiseWindow(display, WindowIds[i]);
  }

}


void StackButtonCB(Widget w_button, XtPointer app_data, XtPointer widget_data)
{
  int i;
  int nb_windows;
  unsigned int rootwidth;
  unsigned int rootheight;
  unsigned int winwidth;
  unsigned int winheight;
  unsigned int xwin;
  unsigned int ywin;
  Display * display;
  Window root;
  unsigned int b_width;
  unsigned int prof;
  int result;

  display = XtDisplay(W_App);

  nb_windows=NbButtons;
  if(nb_windows==0) return;

  /* windows sizes */

  result=XGetGeometry(display, DefaultRootWindow(display), &root, &xwin, &ywin, &rootwidth, &rootheight, &b_width, &prof);

  if(result==0) return;

  winwidth= rootwidth - RWin_LeftOffset - RWin_RightOffset;
  winheight=( (rootheight - RWin_TopOffset - RWin_BottomOffset) - (nb_windows - 1)*RWin_YInterOffset) / nb_windows;

  /* move the windows */

  xwin=RWin_LeftOffset;
  ywin=RWin_TopOffset;

  for(i=0;i<nb_windows;i++) {

    XMoveResizeWindow(display, WindowIds[i], xwin, ywin, winwidth, winheight);
    XRaiseWindow(display, WindowIds[i]);

    ywin+=RWin_YInterOffset + winheight;

  }

}

void CascadeButtonCB(Widget w_button, XtPointer app_data, XtPointer widget_data)
{
  int i;
  int nb_windows;
  unsigned int rootwidth;
  unsigned int rootheight;
  unsigned int winwidth;
  unsigned int winheight;
  unsigned int xwin;
  unsigned int ywin;
  Display * display;
  Window root;
  unsigned int b_width;
  unsigned int prof;
  int result;

  display = XtDisplay(W_App);

  nb_windows=NbButtons;
  if(nb_windows==0) return;

  /* windows sizes */

  result=XGetGeometry(display, DefaultRootWindow(display), &root, &xwin, &ywin, &rootwidth, &rootheight, &b_width, &prof);

  if(result==0) return;

  winwidth= rootwidth - RWin_LeftOffset - RWin_RightOffset - (nb_windows-1)*RWin_CascadeXOffset;
  winheight= rootheight - RWin_TopOffset - RWin_BottomOffset - (nb_windows-1)*RWin_CascadeYOffset;

  /* move the windows */

  xwin=RWin_LeftOffset+(nb_windows-1)*RWin_CascadeXOffset;
  ywin=RWin_TopOffset+(nb_windows-1)*RWin_CascadeYOffset;

  for(i=nb_windows-1;i>=0;i--) {
    XMoveResizeWindow(display, WindowIds[nb_windows-1-i], xwin, ywin, winwidth, winheight);
    XRaiseWindow(display, WindowIds[nb_windows-1-i]);
    xwin-=RWin_CascadeXOffset;
    ywin-=RWin_CascadeYOffset;
  }

}

void QuitOthersButtonCB(Widget w_button, XtPointer app_data, XtPointer widget_data)
{

  int i;
  pid_t current_pid;

  ActivateMenuCB(NULL, NULL, NULL);

  current_pid=getpid();

  for(i=0;i<NbButtons;i++) {
    if(Pids[i]!=current_pid) 
      kill(Pids[i], SIGKILL);
  }

}


void QuitAllButtonCB(Widget w_button, XtPointer app_data, XtPointer widget_data)
{

  int i;
  pid_t current_pid;

  ActivateMenuCB(NULL, NULL, NULL);

  current_pid=getpid();

  for(i=0;i<NbButtons;i++) {
    if(Pids[i]!=current_pid) 
      kill(Pids[i], SIGKILL);
  }

  exit(EXIT_SUCCESS);
}

void InstallWindowMenuCB(void)
{
  Widget w;

  /* Tile Button */

  w = XtNameToWidget(W_App, "*TileButton");
  XtAddCallback(w, XmNactivateCallback, TileButtonCB, NULL);

  /* Stack Button */

  w = XtNameToWidget(W_App, "*StackButton");
  XtAddCallback(w, XmNactivateCallback, StackButtonCB, NULL);

  /* Cascade Button */

  w = XtNameToWidget(W_App, "*CascadeButton");
  XtAddCallback(w, XmNactivateCallback, CascadeButtonCB, NULL);

  /* Quit Others Button */

  w = XtNameToWidget(W_App, "*QuitOthersButton");
  XtAddCallback(w, XmNactivateCallback, QuitOthersButtonCB, NULL);

  /* Quit All Button */

  w = XtNameToWidget(W_App, "*QuitAllButton");
  XtAddCallback(w, XmNactivateCallback, QuitAllButtonCB, NULL);

  /* Window menu */

  w = XtNameToWidget(W_App, "*WindowsMenu");
  XtAddCallback(w, XmNmapCallback, ActivateMenuCB, NULL);
}

void InitModuleWindows(void)
{

  InstallWindowMenuCB();

  GetVarInt("WinLeftOffset", &RWin_LeftOffset);
  GetVarInt("WinTopOffset", &RWin_TopOffset);
  GetVarInt("WinRightOffset", &RWin_RightOffset);
  GetVarInt("WinBottomOffset", &RWin_BottomOffset);
  GetVarInt("WinXInterOffset", &RWin_XInterOffset);
  GetVarInt("WinYInterOffset", &RWin_YInterOffset);
  GetVarInt("WinXCascadeOffset", &RWin_CascadeXOffset);
  GetVarInt("WinYCascadeOffset", &RWin_CascadeYOffset);
}




